#ifndef _NVFS_STRUCT_H
#define _NVFS_STRUCT_H

/*
Layout:
	superblock		1 page
	block bitmap		1 page
	group descriptor	the end is not aligned on page boundary, the rest is inode bitmap
	inode bitmap		the end is not aligned on page boundary, the rest are inodes
	inodes			n pages
	data blocks		m pages
*/

#define DEFAULT_BYTES_PER_INODE	20480

#define NVFS_MAGIC		"NVFS0.03"
#define NVFS_MAGIC_LEN		8

#define NVFS_UUID_LEN		16

#define MIN_LOG_PAGE_SIZE	12
#define MAX_LOG_PAGE_SIZE	16

#define MIN_LOG_INODE_SIZE	8

#define NVFS_DIRECT_BLOCKS	11
#define NVFS_INDIRECT_BLOCKS	5

struct nvfs_sb {
	__u8 magic[NVFS_MAGIC_LEN];
	__u8 uuid[NVFS_UUID_LEN];
	__le64 n_pages;
	__le64 root_inode;
	__u8 log2_page_size;
	__u8 log2_inode_size;
	__u8 log2_inode_pages_per_group;
};

struct nvfs_group {
	__le32 n_free_blocks;
	__le32 n_free_inodes;
	__u8 bitmap[];
};

#define NVFS_XATTR_PREFIX_SHIFT			12
#define NVFS_XATTR_PREFIX_EMPTY			0x0000
#define NVFS_XATTR_PREFIX_USER			0x1000
#define NVFS_XATTR_PREFIX_TRUSTED		0x2000
#define NVFS_XATTR_PREFIX_SECURITY		0x3000
#define NVFS_XATTR_PREFIX_POSIX_ACL_DEFAULT	ACL_TYPE_DEFAULT	/* 0x4000 */
#define NVFS_XATTR_PREFIX_POSIX_ACL_ACCESS	ACL_TYPE_ACCESS		/* 0x8000 */

#define NVFS_XATTR_TOTAL_LEN_MASK		0x000fffff
#define NVFS_XATTR_PREFIX_MASK			0x00f00000
#define NVFS_XATTR_NAMELEN_MASK			0xff000000

#define NVFS_XATTR_TOTAL_LEN_OFFSET		0
#define NVFS_XATTR_PREFIX_OFFSET		20
#define NVFS_XATTR_NAMELEN_OFFSET		24

struct nvfs_xattr_entry {
	__le32 x;
	__u8 data[];
};

static inline unsigned nvfs_xattr_get_total_len(struct nvfs_xattr_entry *e)
{
	return (e->x & NVFS_XATTR_TOTAL_LEN_MASK) >> NVFS_XATTR_TOTAL_LEN_OFFSET;
}

static inline __u8 nvfs_xattr_get_name_len(struct nvfs_xattr_entry *e)
{
	return (e->x & NVFS_XATTR_NAMELEN_MASK) >> NVFS_XATTR_NAMELEN_OFFSET;
}

static inline unsigned nvfs_xattr_get_prefix(struct nvfs_xattr_entry *e)
{
	return (e->x & NVFS_XATTR_PREFIX_MASK) >> (NVFS_XATTR_PREFIX_OFFSET - NVFS_XATTR_PREFIX_SHIFT);
}

static inline size_t nvfs_xattr_entry_needed_size(size_t name_len, size_t data_len)
{
	size_t s = name_len + data_len;
	s = (s + 3) & ~3;
	return sizeof(struct nvfs_xattr_entry) + s;
}

static inline size_t nvfs_xattr_entry_size(const struct nvfs_xattr_entry *nxe)
{
	size_t s = (le32_to_cpu(nxe->x) & NVFS_XATTR_TOTAL_LEN_MASK) >> NVFS_XATTR_TOTAL_LEN_OFFSET;
	s = (s + 3) & ~3;
	return sizeof(struct nvfs_xattr_entry) + s;
}

static inline __le32 nvfs_xattr_entry_make_size(size_t s)
{
	return cpu_to_le32((s - sizeof(struct nvfs_xattr_entry)) << NVFS_XATTR_TOTAL_LEN_OFFSET);
}

struct nvfs_inode {
	__le64 i_size;
	__le64 i_blocks;
	__le64 i_direct_blocks[NVFS_DIRECT_BLOCKS];
	__le64 i_indirect_blocks[NVFS_INDIRECT_BLOCKS];
	__le64 i_xattr_block;
	__le64 i_ctime;
	__le64 i_mtime;
	__le64 i_atime;
	__le32 i_ctime_nsec;
	__le32 i_mtime_nsec;
	__le32 i_atime_nsec;
	__le32 i_uid;
	__le32 i_gid;
	__le32 i_links;
	__le32 i_mode;
	__le32 i_flags;
	struct nvfs_xattr_entry i_xattr[];
};

struct nvfs_dir_entry {
	__le64 inode;
	__le32 name_len_and_type;
	__u8 name[];
};

#define NAME_LEN_MASK	0x00ffffff
#define NAME_LEN_OFFSET	0
#define TYPE_MASK	0xff000000
#define TYPE_OFFSET	24

static inline size_t nvfs_dir_entry_namelen(const struct nvfs_dir_entry *nde)
{
	return (le32_to_cpu(nde->name_len_and_type) & NAME_LEN_MASK) >> NAME_LEN_OFFSET;
}

static inline unsigned char nvfs_dir_entry_type(const struct nvfs_dir_entry *nde)
{
	return (le32_to_cpu(nde->name_len_and_type) & TYPE_MASK) >> TYPE_OFFSET;
}

static inline size_t nvfs_dir_entry_needed_size(size_t name_len)
{
	return (name_len + 19) & ~7;
}

static inline size_t nvfs_dir_entry_size(const struct nvfs_dir_entry *nde)
{
	return nvfs_dir_entry_needed_size(nvfs_dir_entry_namelen(nde));
}

static inline size_t nvfs_dir_entry_make_size(size_t s)
{
	return s - 16;
}


static inline const char *nvfs_validate_super(const struct nvfs_sb *sb, uint64_t size)
{
	uint8_t log2_pages_per_group;
	if (size < 4096)
		return "The device is too small";
	if (memcmp(sb->magic, NVFS_MAGIC, NVFS_MAGIC_LEN))
		return "Bad magic";
	if (sb->log2_page_size < MIN_LOG_PAGE_SIZE || sb->log2_page_size > MAX_LOG_PAGE_SIZE)
		return "Invalid page size";
	if (sb->log2_inode_size < MIN_LOG_INODE_SIZE || sb->log2_inode_size > sb->log2_page_size)
		return "Invalid inode size";

	log2_pages_per_group = sb->log2_page_size + 3;
	if (sb->log2_inode_pages_per_group >= log2_pages_per_group)
		return "Invalid inode pages per group";

	if (size >> sb->log2_page_size < le64_to_cpu(sb->n_pages))
		return "The device is too small";

	return NULL;
}

static inline void nvfs_get_sizes(const struct nvfs_sb *sb,
				  uint64_t *p_n_groups,
				  uint64_t *p_n_inodes,
				  uint64_t *p_inodes_in_group,
				  uint64_t *p_inodes_in_last_group,
				  uint64_t *p_inode_bitmap_inodes)
{
	uint8_t log2_pages_per_group;
	uint64_t n_groups, remaining_pages;
	uint64_t inodes_per_group, inode_bitmap_bytes, inode_bitmap_inodes;
	uint64_t inodes_in_group, inodes_in_last_group;

	log2_pages_per_group = sb->log2_page_size + 3;

	inodes_per_group = 1U << sb->log2_page_size >> sb->log2_inode_size << sb->log2_inode_pages_per_group;
	inode_bitmap_bytes = (inodes_per_group + 7) / 8;
	inode_bitmap_inodes = (sizeof(struct nvfs_group) + inode_bitmap_bytes + ((1U << sb->log2_inode_size) - 1)) >> sb->log2_inode_size;

	n_groups = le64_to_cpu(sb->n_pages) >> log2_pages_per_group;
	remaining_pages = le64_to_cpu(sb->n_pages) & ((1ULL << log2_pages_per_group) - 1);

	inodes_in_group = inodes_per_group - inode_bitmap_inodes;

	if (remaining_pages >= 3) {
		uint64_t i;
		n_groups++;
		inodes_in_last_group = 0;
		for (i = inode_bitmap_inodes; i < inodes_per_group; i++) {
			uint64_t offs = (2 << sb->log2_page_size) + (i << sb->log2_inode_size);
			offs >>= sb->log2_page_size;
			if (offs < remaining_pages)
				inodes_in_last_group++;
			else
				break;
		}
	} else {
		inodes_in_last_group = inodes_in_group;
	}

	if (p_n_groups)
		*p_n_groups = n_groups;
	if (p_n_inodes)
		*p_n_inodes = inodes_in_group * (n_groups - 1) + inodes_in_last_group;
	if (p_inodes_in_group)
		*p_inodes_in_group = inodes_in_group;
	if (p_inodes_in_last_group)
		*p_inodes_in_last_group = inodes_in_last_group;
	if (p_inode_bitmap_inodes)
		*p_inode_bitmap_inodes = inode_bitmap_inodes;
}

#endif
